#!/usr/bin/env python
"""
Image stack -> average -> write FITS
Michael Hirsch

Because ImageJ has been a little buggy about writing FITS files, in particular the header
that astrometry.net then crashes on, we wrote this quick script to ingest a variety
of files and average the specified frames then write a FITS.
The reason we average a selected stack of images is to reduce the noise for use in
astrometry.net

The error you might get from an ImageJ saved FITS when reading in:
PyFits, AstroPy, or ImageMagick is:
IOError: Header missing END card.
"""
from pathlib import Path
from numpy import mean,median,rot90
from astropy.io import fits
from scipy.ndimage import imread
from scipy.io import loadmat
from skimage.color import rgb2gray
import h5py
import tifffile

def meanstack(infn,Navg,ut1=None,method='mean'):
    infn = Path(infn).expanduser()
#%% parse indicies to load
    if isinstance(Navg,slice):
        key = Navg
    elif isinstance(Navg,int):
        key = slice(0,Navg)
    elif len(Navg) == 1:
        key = slice(0,Navg[0])
    elif len(Navg) == 2:
        key = slice(Navg[0],Navg[1])
    else:
        raise ValueError('not sure what you mean by Navg={}'.format(Navg))
#%% load images
    """
    some methods handled individually to improve efficiency with huge files
    """
    if infn.suffix =='.h5':
        with h5py.File(infn,'r',libver='latest') as f:
            img = collapsestack(f['/rawimg'],key,method)
#%% time
            if ut1 is None:
                try:
                    ut1 = f['/ut1_unix'][key][0]
                except KeyError:
                    pass
#%% orientation
            try:
                img = rot90(img,k=f['/params']['rotccw'])
            except KeyError:
                pass

    elif infn.suffix == '.fits':
        with fits.open(infn,mode='readonly',memmap=False) as f: #mmap doesn't work with BZERO/BSCALE/BLANK
            img = collapsestack(f[0].data, key,method)

    elif infn.suffix.startswith('.tif'):
        img = tifffile.imread(str(infn),key=key)
        img = collapsestack(img,key,method)

    elif infn.suffix == '.mat':
        img = loadmat(infn)
        img = collapsestack(img['data'].T, key, method) #matlab is fortran order
    else:
        img = imread(infn)
        if img.ndim in (3,4) and img.shape[-1]==3: #assume RGB
            img = collapsestack(rgb2gray(img),key,method)

    return img,ut1

def collapsestack(img,key,method):
    if img.ndim==2:
        return img
#%%
    if img.ndim==3:
        if method=='mean':
            method=mean
        elif method=='median':
            method=median
        else:
            raise TypeError('unknown method {}'.format(method))

        return method(img[key,...],axis=0).astype(img.dtype)

def writefits(img,outfn):
    outfn = Path(outfn).expanduser()
    print(f'writing {outfn}')

    f = fits.PrimaryHDU(img)
    f.writeto(outfn, clobber=True,checksum=True)
    #no close

if __name__ == '__main__':
    from argparse import ArgumentParser
    p = ArgumentParser(description='takes mean of first N images and writes new FITS file')
    p.add_argument('infn',help='FITS input file to read')
    p.add_argument('-N','--Navg',help='avg first N frames, or range(start,stop) of frame if two args given',nargs='+',default=(10,),type=int)
    p = p.parse_args()

    infn = Path(p.infn)

    meanimg = meanstack(infn,p.Navg)

    writefits(meanimg,infn.with_suffix('_mean.fits'))



from pathlib import Path
import subprocess
import h5py
from numpy import meshgrid,column_stack
from astropy.wcs import wcs
import logging
from astropy.io import fits #instead of obsolete pyfits
from dateutil.parser import parse
from datetime import datetime
from pytz import UTC
#
from pymap3d import radec2azel
#
from .plots import plotazel,plotradec
#%%

def readfitsimagestack(fn):
    fn = Path(fn).expanduser()

    with fits.open(fn, mode='readonly') as f:
        if len(f[0].shape) > 1: #no .ndim in Astropy.io.fits 1.2
            return f[0].data
        else:
            raise TypeError(f'did not find image data in {fn}. This may be a bug.')


def fits2radec(fitsfn,camLatLon,makeplot,clim=None):
    fitsfn = Path(fitsfn).expanduser()

    if fitsfn.suffix == '.fits':
        WCSfn = fitsfn.with_suffix('.wcs') #using .wcs will also work but gives a spurious warning

    elif fitsfn.suffix == '.wcs':
        makeplot.append('skipsolve')
        WCSfn = fitsfn
    else:
        raise TypeError(f'please convert {fitsfn} to GRAYSCALE .fits e.g. with ImageJ or ImageMagick')

    if not 'skipsolve' in makeplot:
        doSolve(fitsfn)

    with fits.open(fitsfn, mode='readonly') as f:
        yPix,xPix = f[0].shape[-2:]

    x,y = meshgrid(range(xPix), range(yPix)) #pixel indices to find RA/dec of
    xy = column_stack((x.ravel(order='C'), y.ravel(order='C')))
#%% use astropy.wcs to register pixels to RA/DEC
    """
    http://docs.astropy.org/en/stable/api/astropy.wcs.WCS.html#astropy.wcs.WCS
    naxis=[0,1] is to take x,y axes in case a color photo was input e.g. to astrometry.net cloud solver
    """
    try:
        with fits.open(WCSfn, mode='readonly') as f:
            #radec = wcs.WCS(hdul[0].header,naxis=[0,1]).all_pix2world(xy, 0)
            radec = wcs.WCS(f[0].header).all_pix2world(xy, 0)
    except IOError:
        raise RuntimeError(f'It appears the WCS solution is not present, was the FITS image solved?  looking for: {WCSfn}')

    ra  = radec[:,0].reshape((yPix,xPix),order='C')
    dec = radec[:,1].reshape((yPix,xPix),order='C')
#%% plot
    plotradec(ra,dec,x,y,camLatLon,fitsfn,makeplot)

    return x,y,ra,dec


def r2ae(fitsFN,ra,dec,x,y,camLatLon,specTime,makeplot):
    if camLatLon is not None and ra is not None:
        try:
            if specTime is None:
                with fits.open(fitsFN,mode='readonly') as f:
                    frameTime = f[0].header['FRAME'] #TODO this only works from Solis?
                    timeFrame = parse(frameTime).replace(tzinfo=UTC)
                    logging.warning('assumed UTC time zone, using FITS header for time')
            elif isinstance(specTime,datetime):
                timeFrame = specTime
            elif isinstance(specTime,(float,int)): #assume UT1_Unix
                timeFrame = datetime.fromtimestamp(specTime,tz=UTC)
            else: #user override of frame time
                timeFrame = parse(specTime)
        except (KeyError, TypeError):
            logging.error(f'could not read time from {fitsFN}')
            return (None,)*3
    else:
        return (None,)*3

    print(f'image time: {timeFrame}')

#%% knowing camera location, time, and sky coordinates observed, convert to az/el for each pixel
    az,el = radec2azel(ra, dec, camLatLon[0], camLatLon[1],timeFrame)

    plotazel(az,el,x,y,fitsFN,camLatLon,timeFrame,makeplot)

    return az,el,timeFrame


def doSolve(fitsfn):
    """
    Astrometry.net from at least version 0.67 is OK with Python 3.
    """
    #binpath = Path(find_executable('solve-field')).parent

    cmd = ['solve-field','--overwrite', str(fitsfn)]
    print(cmd)

    subprocess.run(cmd)#, env={'PATH':str(binpath)})

    print('\n\n *** done with astrometry.net ***\n ')


def fits2azel(fitsfn,camLatLon,specTime,makeplot, clim=None):
    fitsfn = Path(fitsfn).expanduser()

    x,y,ra,dec = fits2radec(fitsfn,camLatLon,makeplot, clim)
    az,el,timeFrame = r2ae(fitsfn,ra,dec,x,y,camLatLon,specTime, makeplot)

    if 'h5' in makeplot:
        h5fn = fitsfn.with_suffix('.h5')
        print(f'saving {h5fn}')
        writeh5(h5fn,az,el,camLatLon,x,y,ra,dec,timeFrame,fitsfn)

    return x,y,ra,dec,az,el,timeFrame


def writeh5(h5fn,az,el,camLatLon,x,y,ra,dec,timeFrame,fitsfn):
    with h5py.File(h5fn, 'w',libver='latest') as f:

        if az is not None:
            h5az = f.create_dataset("/az",data=az,compression="gzip",shuffle=True,fletcher32=True)
            h5az.attrs['Units'] = 'degrees'

            h5el = f.create_dataset("/el",data=el,compression="gzip",shuffle=True,fletcher32=True)
            h5el.attrs['Units'] = 'degrees'

            h5ll = f.create_dataset("/sensorloc",data=camLatLon,shuffle=True,fletcher32=True)
            h5ll.attrs['Units'] = 'WGS-84 degrees'

        h5x =  f.create_dataset("/x",data=x,compression="gzip",shuffle=True,fletcher32=True)
        h5x.attrs['Units'] = 'pixels'

        h5y =  f.create_dataset("/y",data=y,compression="gzip",shuffle=True,fletcher32=True)
        h5y.attrs['Units'] = 'pixels'

        h5ra = f.create_dataset("/ra",data=ra,compression="gzip",shuffle=True,fletcher32=True)
        h5ra.attrs['Units'] = 'degrees'

        h5dec =f.create_dataset("/dec",data=dec,compression="gzip",shuffle=True,fletcher32=True)
        h5dec.attrs['Units'] = 'degrees'

        h5tf = f.create_dataset("/timeFrame",data=str(timeFrame))
        h5tf.attrs['Units'] = 'UTC time of exposure start'

        h5fn = f.create_dataset('/fitsfn',data=str(fitsfn))
        h5fn.attrs['Units'] = 'original FITS filename'



#!/usr/bin/env python
from setuptools import setup

req = ['pymap3d',
        'nose','pillow','python-dateutil','pytz','numpy','scipy','h5py','astropy','scikit-image','matplotlib','seaborn' ]

setup(name='astrometry_azel',
      packages=['astrometry_azel'],
      author='Michael Hirsch, Ph.D.',
      description='Register images to az/el using the astrometry.net program',
      url='https://github.com/scivision/astrometry_azel',
      version='1.1.0',
      classifiers=[
      'Intended Audience :: Science/Research',
      'Development Status :: 5 - Production/Stable',
      'License :: OSI Approved :: MIT License',
      'Topic :: Scientific/Engineering :: Atmospheric Science',
      'Programming Language :: Python :: 3.6',
      ],
      install_requires=req,
      extras_require={'pyfits':'pyfits'},
	  )




from pathlib import Path
import logging
from matplotlib.pyplot import figure

def plotazel(az,el,x=None,y=None, fn='',camLatLon='',timeFrame='',makeplot='',ttxt=''):
    fn = Path(fn)

    if 'show' in makeplot or 'png' in makeplot:
        plottype = 'contour'

        fg = figure(figsize=(12,5))
        axa = fg.add_subplot(1,2,1)
        if plottype=='image':
            hia = axa.imshow(az,origin='lower')
            hc = fg.colorbar(hia)
            hc.set_label('Azimuth [deg]')
        elif plottype == 'contour':
            if x is not None and y is not None:
                cs = axa.contour(x,y,az)
            else:
                cs = axa.contour(az)
            axa.clabel(cs, inline=1,fmt='%0.1f')

        axa.set_xlabel('x-pixel')
        axa.set_ylabel('y-pixel')

        ttxt += 'az.: {} {} {}'.format(fn.name, camLatLon, timeFrame)
        axa.set_title(ttxt)

        axe = fg.add_subplot(1,2,2)
        if plottype=='image':
            hie = axe.imshow(el,origin='lower')
            hc = fg.colorbar(hie)
            hc.set_label('Elevation [deg]')
        elif plottype == 'contour':
            if x is not None and y is not None:
                cs = axe.contour(x,y,el)
            else:
                cs = axe.contour(el)
            axe.clabel(cs, inline=1,fmt='%0.1f')
        axe.set_xlabel('x-pixel')
        axe.set_ylabel('y-pixel')
        axe.set_title('el.: {} {} {}'.format(fn.name, camLatLon, timeFrame))
#%%
        if 'png' in makeplot:
            plotFN = fn.parent/(fn.stem+'_azel.png')
            print(f'writing  {plotFN}')
            fg.savefig(str(plotFN), bbox_inches='tight',dpi=150)

        return fg,axa,axe

def plotradec(ra,dec,x,y,camLatLon,fn,makeplot):
    fn = Path(fn).expanduser()

    ttxt = f'{fn.name}'
    if camLatLon:
        ttxt += str(camLatLon)

    if 'show' in makeplot or 'png' in makeplot:
        plottype = 'contour'

        fg = figure(figsize=(12,5))
        ax = fg.add_subplot(1,2,1)
        if plottype=='image':
            hri = ax.imshow(ra,origin='lower')
            hc = fg.colorbar(hri)
            hc.set_label('RA [deg]')
        elif plottype == 'contour':
            cs= ax.contour(x,y,ra)
            ax.clabel(cs, inline=1,fmt='%0.1f')

        ax.set_xlabel('x-pixel')
        ax.set_ylabel('y-pixel')
        ax.set_title('RA: ' + ttxt)
#%%
        ax = fg.add_subplot(1,2,2)
        if plottype=='image':
            hdi = ax.imshow(dec,origin='lower')
            hc = fg.colorbar(hdi)
            hc.set_label('Dec [deg]')
        elif plottype == 'contour':
            cs= ax.contour(x,y,dec)
            ax.clabel(cs, inline=1,fmt='%0.1f')
        ax.set_xlabel('x-pixel')
        ax.set_ylabel('y-pixel')
        ax.set_title('DECL: ' + ttxt)
#%%
        if 'png' in makeplot:
            plotFN = fn.parent/(fn.stem+'_radec.png')
            print(f'writing  {plotFN}')
            fg.savefig(str(plotFN), bbox_inches='tight', dpi=150)

def plotimagestack(img,fn,makeplot,clim=None):
    fn = Path(fn).expanduser()
    #%% plotting
    if img.ndim==3 and img.shape[0] == 3: #it seems to be an RGB image
        cmap = None
        imnorm = None
        img = img.transpose([1,2,0]) #imshow() needs colors to be last axis
    else:
        cmap = 'gray'
        imnorm=None
        #imnorm = LogNorm()

    fg = figure()
    ax = fg.gca()
    if clim is None:
        hi = ax.imshow(img,origin='lower',interpolation='none',cmap=cmap,norm=imnorm)
    else:
        hi = ax.imshow(img,origin='lower',interpolation='none',cmap=cmap,vmin=clim[0],vmax=clim[1],norm=imnorm)
    ax.set_xlabel('x')
    ax.set_ylabel('y')
    ax.set_title(str(fn))
    if cmap is not None:
        try:
            hc = fg.colorbar(hi)
            hc.set_label(f'Data numbers {img.dtype}')
        except Exception as e:
            logging.warning(f'trouble making picture colorbar  {e}')
#%%
    if 'png' in makeplot:
        plotFN = fn.parent/(fn.stem+'_picture.png')
        print(f'writing  {plotFN}')
        fg.savefig(str(plotFN), bbox_inches='tight', dpi=150)
